//
//  NoteManager.m
//  ForClark
//
//  Created by Clark Barry on 12/21/12.
//  Copyright (c) 2012 CSHaus. All rights reserved.
//

#import "NoteManager.h"
#import "LittleNoteParser.h"
#import "URLFeeder.h"
#import "Note.h"
#import "User.h"
#import "CommonDefines.h"

NSString *kNewNotesNotification = @"new notes notification";
NSString *kNoNewNotesNotification = @"no new notes";
NSString *kNoteSubmitFinishedNotification = @"note submit";
NSString *kNoteSubmitFailedNotification = @"note submit failed";

@interface NoteManager (Internal)
-(void)loadNotes;

+(NSString *)getDocumentsDirectory;
+(Note *)noteForPath:(NSString *)path;
+(NSString *)pathForNoteIndex:(NSUInteger)index;
@end

@implementation NoteManager

@synthesize notes = _notes;
@synthesize attemptingFetch = _attemptingFetch;

static NoteManager *__instance = nil;



+(NoteManager *)getInstance{
    if(!__instance){
        static dispatch_once_t onceToken;
        dispatch_once(&onceToken, ^{
            __instance = [[NoteManager alloc] init];
        });
    }
    return __instance;
}

-(id)init{
    self = [super init];
    if(self){
        _attemptingFetch = NO;
        [self loadNotes];
    }
    return self;
}

-(void)fetchNotes{
    _attemptingFetch = YES;
    [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
    NSString *url = [URLFeeder getURLForNotesJson];
    url = [[User getInstance]addAuthToken:url];
    Note *lastNote = [_notes lastObject];
    unsigned int lastIndex = [lastNote getIndex];
    url = [url stringByAppendingFormat:@"&since_id=%d",lastIndex];
    NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:url]];
    [request setHTTPMethod:@"GET"];
    [NSURLConnection sendAsynchronousRequest:request
                                       queue:[[NSOperationQueue alloc]init]
                           completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                               if(error){
                                   [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                            object:nil]];
                               }else if(data){
                                   NSHTTPURLResponse *responseHTTP = (NSHTTPURLResponse *)response;
                                   if([responseHTTP statusCode] == 200){
                                       NSError *jsonError;
                                       NSArray *jsonArray = [NSJSONSerialization JSONObjectWithData:data options:kNilOptions error:&jsonError];
                                       NSArray *newNotes = [[LittleNoteParser getInstance] getNotesFromArrayOfJSON:jsonArray];
                                       if([newNotes count]){
                                           [self addNotes:newNotes];
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kNewNotesNotification
                                                                                                                               object:nil]];
                                       }else{
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kNoNewNotesNotification
                                                                                                                               object:nil]];
                                       }
                                   }
                               }else{
                                   [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kNoNewNotesNotification
                                                                                                                       object:nil]];
                               }
                                _attemptingFetch = NO;
                                dispatch_async(dispatch_get_main_queue(), ^{
                                    [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                });

                            }];
}

-(void)submitNote:(Note *)note{
    if(![[User getInstance]getCSRFToken]){
        [[User getInstance]getCookie:^{
            [self submitNote:note];
        }];
    }else{
        _attemptingFetch = YES;
        [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
        NSString *url = [URLFeeder getURLForNotesJson];
        url = [[User getInstance]addAuthToken:url];
        Note *lastNote = [_notes lastObject];
        unsigned int lastIndex = [lastNote getIndex];
        url = [url stringByAppendingFormat:@"&since_id=%d",lastIndex];
        NSURL *notesURL = [NSURL URLWithString:url];
        NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:notesURL];
        [request setValue:[[User getInstance] getCSRFToken] forHTTPHeaderField:@"X-csrf-token"];
        [request setValue:@"authenticity_token" forHTTPHeaderField:@"X-csrf-param"];
        [request setHTTPMethod:@"POST"];
        NSString *body = [NSString stringWithFormat:@"note[%@]=%@",CONTENT_KEY, note.text];

        [request setHTTPBody:[body dataUsingEncoding:NSUTF8StringEncoding]];
        [NSURLConnection sendAsynchronousRequest:request
                                           queue:[[NSOperationQueue alloc]init]
                               completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                                   if(error){
                                       [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                            object:nil]];
                                   }else if(data){
                                       NSHTTPURLResponse *responseHTTP = (NSHTTPURLResponse *)response;
                                       if([responseHTTP statusCode] == 201){
                                           NSError *error;
                                           NSArray *json = [NSJSONSerialization JSONObjectWithData:data options:kNilOptions error:&error];
                                           NSArray *newNotes = [[LittleNoteParser getInstance] getNotesFromArrayOfJSON:json];
                                           if([newNotes count]){
                                               [self addNotes:newNotes];
                                               [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kNewNotesNotification
                                                                                                                                   object:nil]];
                                           }else{
                                               [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kNoNewNotesNotification
                                                                                                                                   object:nil]];
                                           }
                                           [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kNoteSubmitFinishedNotification
                                                                                                                                object:nil]];
                                       }else{
                                           [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kNoteSubmitFailedNotification
                                                                                                                                object:nil]];
                                       }
                                       dispatch_async(dispatch_get_main_queue(), ^{
                                           [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                       });
                                   }
                               }];
    }
}

+(Note *)noteAtIndex:(NSUInteger)index{
    return [[[NoteManager getInstance]notes] objectAtIndex:index];
}

+(NSUInteger)getNoteCount{
    return [[[NoteManager getInstance]notes] count];
}

-(NSArray *)notes{
    if(!_notes){
        [self loadNotes];
    }
    return _notes;
}

+(NSString *)getDocumentsDirectory{
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES);
    NSString *documentsDirectory = [paths objectAtIndex:0];
    documentsDirectory = [documentsDirectory stringByAppendingPathComponent:@"Notes"];
    
    NSError *error;
    [[NSFileManager defaultManager] createDirectoryAtPath:documentsDirectory withIntermediateDirectories:YES attributes:nil error:&error];
    
    return documentsDirectory;
}

+(Note *)noteForPath:(NSString *)path{
    path = [path stringByAppendingPathComponent:DATA_FILE];
    NSData *data = [[NSData alloc] initWithContentsOfFile:path];
    NSKeyedUnarchiver *reader = [[NSKeyedUnarchiver alloc] initForReadingWithData:data];
    Note *note = [reader decodeObjectForKey:DATA_KEY];
    return note;
}

+(NSString *)pathForNoteIndex:(NSUInteger)index{
    NSString *path = [[self class] getDocumentsDirectory];
    path = [path stringByAppendingPathComponent:[NSString stringWithFormat:@"%04d.note",index]];
    return path;
}

-(void)loadNotes{
    NSString *dataDirectory = [[self class]getDocumentsDirectory];
    NSError *error;
    NSArray *files = [[NSFileManager defaultManager]contentsOfDirectoryAtPath:dataDirectory error:&error];
    NSMutableArray *tempNotes = [NSMutableArray arrayWithCapacity:[files count]];
    for(NSString *file in files){
        NSString *fullPath = [dataDirectory stringByAppendingPathComponent:file];
        Note *note = [[self class] noteForPath:fullPath];
        [tempNotes addObject:note];
    }
    _notes = [NSMutableArray arrayWithArray:tempNotes];
}

-(void)addNotes:(NSArray *)newNotes{
    NSUInteger index = [_notes count];
    [_notes addObjectsFromArray:newNotes];
    NSString *filePath;
    for(Note *note in newNotes){
        filePath = [[self class]pathForNoteIndex:index];
        NSError *error;
        if(![[NSFileManager defaultManager] createDirectoryAtPath:filePath withIntermediateDirectories:YES attributes:nil error:&error]){
            NSLog(@"Error savingToNotes: %@",[error localizedDescription]);
            return;
        }
        filePath = [filePath stringByAppendingPathComponent:DATA_FILE];
        
        NSMutableData *noteData = [[NSMutableData alloc]init];
        NSKeyedArchiver *writer = [[NSKeyedArchiver alloc]initForWritingWithMutableData:noteData];
        [writer encodeObject:note forKey:DATA_KEY];
        [writer finishEncoding];
        [noteData writeToFile:filePath atomically:YES];
        index++;
    }
}

-(void)deleteAll{
    _notes = nil;
    NSFileManager *fm = [NSFileManager defaultManager];
    NSString *directory = [[self class] getDocumentsDirectory];
    NSError *error = nil;
    for (NSString *file in [fm contentsOfDirectoryAtPath:directory error:&error]) {
        NSString *fullPath = [directory stringByAppendingPathComponent:file];
        BOOL success = [fm removeItemAtPath:fullPath error:&error];
        if (!success || error) {
            // it failed.
        }
    }
}


@end
