//
//  User.m
//  LittleNote
//
//  Created by Clark Barry on 1/18/13.
//  Copyright (c) 2013 CSHaus. All rights reserved.
//

#import "User.h"

#import "URLFeeder.h"
#import "CommonDefines.h"
#import "LittleNoteParser.h"
#import "NoteManager.h"
#import "Request.h"
#import "KeychainItemWrapper.h"

NSString *kLoginSuccessfulNotification = @"log in successful";
NSString *kAuthenticaitonFailedNotification =  @"authentication failed";

NSString *kUserCreateSuccessfulNotification = @"create user successful";
NSString *kUserCreateFailedNotification = @"create user failed";

NSString *kConnectionFailedNotification = @"connection failed";

NSString *kUserLogoutNotification = @"logout";
NSString *kNewRequestNotification = @"new request";
NSString *kRequestCancelledNotification = @"request cancelled";
NSString *kUserDoesNotExistNotification = @"no such user";
NSString *kRequestResponseSuccessful = @"request response";
NSString *kStopSharingSuccussfulNotification = @"stopped sharing";

@interface User (Internal)
-(void)getCookie:(void(^)(void))completion;
-(NSMutableURLRequest *)getLittleNoteRequestForURL:(NSURL *)url;
@end

@implementation User

static User *__instance;
+(User *)getInstance{
    if(!__instance){
        static dispatch_once_t onceToken;
        dispatch_once(&onceToken, ^{
            __instance = [[User alloc]init];
        });
    }
    return __instance;
}

-(id)init{
    self = [super init];
    if(self){
        KeychainItemWrapper *keychain = [[KeychainItemWrapper alloc] initWithIdentifier:LITTLE_NOTE_IDENTIFIER accessGroup:nil];
        _username = [keychain objectForKey:((__bridge id)kSecAttrAccount)];
        _authToken = [keychain objectForKey:((__bridge id)kSecValueData)];
        _personSharingWith = [[NSUserDefaults standardUserDefaults] objectForKey:SHARE_WITH_KEY];
    }
    return self;
}

-(void)getCookie:(void (^)(void))completion{
    NSURL *url = [NSURL URLWithString:[URLFeeder getCookie]];
    NSURLRequest *request = [NSURLRequest requestWithURL:url];
    [NSURLConnection sendAsynchronousRequest:request
                                       queue:[[NSOperationQueue alloc]init]
                           completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                               if(error){
                                   [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                       object:nil]];
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                       [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                   });
                               }else if(data){
                                   NSError *jsonError;
                                   NSDictionary *results = [NSJSONSerialization JSONObjectWithData:data options:kNilOptions error:&jsonError];
                                   NSString *tempString = [results objectForKey:CSRF_KEY];
                                   _csrfToken = tempString;
                                   completion();
                               }
                           }];
}

-(void)setAuthToken:(NSString *)authToken{
    _authToken = authToken;
}

-(NSString *)addAuthToken:(NSString *)body{
    return [body stringByAppendingFormat:@"?auth_token=%@",_authToken];
}

-(BOOL)hasCredentials{
    return _username && _authToken && ![_username isEqualToString:@""] && ![_authToken isEqualToString:@""];
}

-(NSString *)getCSRFToken{
    return _csrfToken;
}


-(NSMutableURLRequest *)getLittleNoteRequestForURL:(NSURL *)url{
    NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url];
    [request setTimeoutInterval:15];
    [request setValue:_csrfToken forHTTPHeaderField:@"X-csrf-token"];
    return request;
}


-(void)attemptLoginWithUsername:(NSString *)username andPassword:(NSString *)password{
    dispatch_async(dispatch_get_main_queue(), ^{
        [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
    });
    if(!_csrfToken){
        [self getCookie:^{
            [self attemptLoginWithUsername:username andPassword:password];
        }];
    }else{
        NSURL *notesURL = [NSURL URLWithString:[URLFeeder getLoginURL]];
        NSMutableURLRequest *request = [self getLittleNoteRequestForURL:notesURL];
        NSString *userField = [NSString stringWithFormat:@"user[login]=%@",username];
        NSString *passwordField = [NSString stringWithFormat:@"user[password]=%@",password];
        [request setHTTPMethod:@"POST"];
        NSString *body = [NSString stringWithFormat:@"%@&%@",userField,passwordField];
        [request setHTTPBody:[body dataUsingEncoding:NSUTF8StringEncoding]];
        [NSURLConnection sendAsynchronousRequest:request
                                           queue:[[NSOperationQueue alloc]init]
                               completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                                   if(error){
                                       NSLog(@"error = %@",[error localizedDescription]);
                                       [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                           object:nil]];
                                    }else{
                                        NSError *jsonError;
                                        NSDictionary *results = [NSJSONSerialization JSONObjectWithData:data options:kNilOptions error:&jsonError];
                                        NSHTTPURLResponse *responseHTTP = (NSHTTPURLResponse *)response;
                                        if([responseHTTP statusCode] == 200){
                                            NSString *success = [results objectForKey:LOGIN_SUCCESS_KEY];
                                            if(![success boolValue]){
                                                [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kAuthenticaitonFailedNotification
                                                                                                                                    object:nil]];
                                            }else{
                                                _username = username;
                                                NSString *authToken = [results objectForKey:AUTH_TOKEN_KEY];
                                                _authToken = authToken;
                                                KeychainItemWrapper *keychain = [[KeychainItemWrapper alloc] initWithIdentifier:LITTLE_NOTE_IDENTIFIER accessGroup:nil];
                                                [keychain setObject:_username forKey:((__bridge id)kSecAttrAccount)];
                                                [keychain setObject:_authToken forKey:((__bridge id)kSecValueData)];
                                            
                                                NSDictionary *user = [results objectForKey:USER_KEY];

                                                NSArray *notesJson = [user objectForKey:NOTES_KEY];
                                                NSArray *notes = [[LittleNoteParser getInstance] getNotesFromArrayOfJSON:notesJson];
                                                NSInteger currentNoteCount = [NoteManager getNoteCount];

                                                if([notes count] > currentNoteCount){
                                                    [[NoteManager getInstance]addNotes:notes];
                                                }
                                                
                                                NSDictionary *requestsJson = [user objectForKey:REQUEST_KEY];
                                                NSArray *requestsTo = [requestsJson objectForKey:REQUEST_TO_KEY];
                                                _requests = [NSMutableArray arrayWithArray:[[LittleNoteParser getInstance]getRequestsFromArrayOfJSON:requestsTo]];
                                                
                                                NSArray *pendingNames = [requestsJson objectForKey:REQUEST_PENDING_KEY];
                                                if([pendingNames count]){
                                                    NSDictionary *pendingRequest = [pendingNames objectAtIndex:0];
                                                    NSDictionary *toInfo = [pendingRequest objectForKey:REQUEST_TO_KEY];
                                                    NSString *name = [toInfo objectForKey:NAME_KEY];
                                                    _personSharingWith = name;
                                                }
                                                
                                                NSDictionary *share = [user objectForKey:SHARE_WITH_KEY];
                                                NSString *sharingWith = [share objectForKey:NAME_KEY];
                                                if(sharingWith){
                                                    _personSharingWith = [NSString stringWithString:sharingWith];
                                                    [[NSUserDefaults standardUserDefaults]setObject:_personSharingWith forKey:SHARE_WITH_KEY];
                                                    [[NSUserDefaults standardUserDefaults] synchronize];
                                                }
                                                
                                                [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kLoginSuccessfulNotification
                                                                                                                                object:nil]];
                                            }
                                       }else{
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kAuthenticaitonFailedNotification
                                                                                                                               object:nil]];
                                       }
                                    }
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                       [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                   });
                                }];
    }
}

-(void)attemptCreateUserWith:(NSString *)username Password:(NSString *)password andEmail:(NSString *)email{
    dispatch_async(dispatch_get_main_queue(), ^{
        [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
    });
    if(!_csrfToken){
        [self getCookie:^{
            [self attemptCreateUserWith:username Password:password andEmail:email];
        }];
    }else{
        NSURL *notesURL = [NSURL URLWithString:[URLFeeder getURLForUserCreate]];
        NSMutableURLRequest *request = [self getLittleNoteRequestForURL:notesURL];
        [request setHTTPMethod:@"POST"];
        NSString *user = [NSString stringWithFormat:@"user[name]=%@",username];
        NSString *passwordField = [NSString stringWithFormat:@"user[password]=%@",password];
        NSString *passwordConfirmation = [NSString stringWithFormat:@"user[password_confirmation]=%@",password];
        NSString *emailField = [NSString stringWithFormat:@"user[email]=%@",email];
        NSString *body = [NSString stringWithFormat:@"%@&%@&%@&%@",user,passwordField,passwordConfirmation,emailField];
        [request setHTTPBody:[body dataUsingEncoding:NSUTF8StringEncoding]];
        [NSURLConnection sendAsynchronousRequest:request
                                           queue:[[NSOperationQueue alloc]init]
                               completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                                   if(error){
                                       [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                           object:nil]];
                                   }else{
                                       NSHTTPURLResponse *httpRespone = (NSHTTPURLResponse *)response;
                                       NSInteger status = [httpRespone statusCode];
                                       if(status == 200){
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kUserCreateSuccessfulNotification
                                                                                                                               object:nil]];
                                       }else if(status == 422){
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kUserCreateFailedNotification
                                                                                                                               object:nil]];
                                       }
                                   }
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                       [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                   });
                               }];
    }
}

-(void)requestShareWithUser:(NSString *)targetName{
    dispatch_async(dispatch_get_main_queue(), ^{
        [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
    });
    if(!_csrfToken){
        [self getCookie:^{
            [self requestShareWithUser:targetName];
        }];
    }else{
        NSURL *url = [NSURL URLWithString:[self addAuthToken:[URLFeeder getURLForRequest]]];
        NSMutableURLRequest *request = [self getLittleNoteRequestForURL:url];
        [request setHTTPMethod:@"POST"];
        NSString *userField = [NSString stringWithFormat:@"from=%@",_username];
        NSString *targetField = [NSString stringWithFormat:@"to=%@",targetName];
        NSString *body = [NSString stringWithFormat:@"%@&%@",userField,targetField];
        [request setHTTPBody:[body dataUsingEncoding:NSUTF8StringEncoding]];
        [NSURLConnection sendAsynchronousRequest:request
                                           queue:[[NSOperationQueue alloc]init]
                               completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                                   if(error){
                                       [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                           object:nil]];
                                   }else{
                                       NSHTTPURLResponse *httpRespone = (NSHTTPURLResponse *)response;
                                       if([httpRespone statusCode]==201){
                                           _personSharingWith = targetName;
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kRequestResponseSuccessful
                                                                                                                               object:nil]];
                                       }else if([httpRespone statusCode]==404){
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kUserDoesNotExistNotification
                                                                                                                               object:nil]];
                                       }
                                   }
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                       [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                   });
                               }];
    }
}

-(void)checkRequests{
    dispatch_async(dispatch_get_main_queue(), ^{
        [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
    });
    if(!_csrfToken){
        [self getCookie:^{
            [self checkRequests];
        }];
    }else{
        NSString *urlString = [self addAuthToken:[URLFeeder getURLForRequest]];
        NSURL *url = [NSURL URLWithString:urlString];
        NSMutableURLRequest *request = [self getLittleNoteRequestForURL:url];
        [request setHTTPMethod:@"GET"];
        [NSURLConnection sendAsynchronousRequest:request
                                           queue:[[NSOperationQueue alloc]init]
                               completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                                   if(error){
                                       [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                       object:nil]];
                                   }else{
                                       NSHTTPURLResponse *httpRespone = (NSHTTPURLResponse *)response;
                                       if([httpRespone statusCode] == 200){
                                           NSError *jsonError;
                                           NSDictionary *allRequests = [NSJSONSerialization JSONObjectWithData:data options:kNilOptions error:&jsonError];
                                           NSArray *pendingJson = [allRequests objectForKey:REQUEST_TO_KEY];
                                           NSUInteger requestCount = _requests ? [_requests count] : 0;
                                           _requests = [NSMutableArray arrayWithArray:[[LittleNoteParser getInstance]getRequestsFromArrayOfJSON:pendingJson]];
                                           
                                           NSArray *pendingNames = [allRequests objectForKey:REQUEST_PENDING_KEY];
                                           if([pendingNames count]){
                                               NSDictionary *pendingRequest = [pendingNames objectAtIndex:0];
                                               NSDictionary *toInfo = [pendingRequest objectForKey:REQUEST_TO_KEY];
                                               NSString *name = [toInfo objectForKey:NAME_KEY];
                                               _personSharingWith = name;
                                           }else if(requestCount < [_requests count] && !_personSharingWith){
                                               [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kNewRequestNotification
                                                                                                                                    object:nil]];
                                           }
                                       }
                                   }
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                       [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                   });
                               }];

    }
}

-(BOOL)hasRequests{
    return [_requests count] && !_personSharingWith;
}

-(Request *)firstRequest{
    if([_requests count]){
        return [_requests objectAtIndex:0];
    }else{
        return nil;
    }
}
-(void)respondToRequest:(Request *)r withResponse:(BOOL)accepted{
    dispatch_async(dispatch_get_main_queue(), ^{
        [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
    });
    if(!_csrfToken){
        [self getCookie:^{
            [self respondToRequest:r withResponse:accepted];
        }];
    }else{
        NSString *urlString = [self addAuthToken:[URLFeeder getURLForRequestResponse:r]];
        urlString = [urlString stringByAppendingString:@"&accept="];
        NSString *responseString = accepted ? @"true" : @"false";
        urlString = [urlString stringByAppendingString:responseString];
        NSURL *url = [NSURL URLWithString:urlString];
        NSMutableURLRequest *request = [self getLittleNoteRequestForURL:url];
        [request setHTTPMethod:@"PUT"];
        [NSURLConnection sendAsynchronousRequest:request
                                           queue:[[NSOperationQueue alloc]init]
                               completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
                                   if(error){
                                       [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                           object:nil]];
                                   }else{
                                       NSHTTPURLResponse *httpRespone = (NSHTTPURLResponse *)response;
                                       if([httpRespone statusCode] == 204){
                                           if(accepted){
                                               _personSharingWith = r.requesterName;
                                               [[NSUserDefaults standardUserDefaults]setObject:_personSharingWith forKey:SHARE_WITH_KEY];
                                               [[NSUserDefaults standardUserDefaults] synchronize];
                                           }
                                           [_requests removeObjectAtIndex:0];
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kRequestResponseSuccessful
                                                                                                                               object:nil]];
                                       }else if([httpRespone statusCode] == 404){
                                           [_requests removeObjectAtIndex:0];
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kRequestCancelledNotification
                                                                                                                               object:nil]];
                                       }
                                   }
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                       [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                   });
                               }];
        
    }
}

-(void)logout{
    _authToken = nil;
    _username = nil;
    _personSharingWith = nil;
    _requests = nil;
    [[NoteManager getInstance] deleteAll];
    NSHTTPCookie *cookie;
    NSHTTPCookieStorage *storage = [NSHTTPCookieStorage sharedHTTPCookieStorage];
    for (cookie in [storage cookies]) {
        [storage deleteCookie:cookie];
    }
    [[NSUserDefaults standardUserDefaults] synchronize];
    KeychainItemWrapper *keychain = [[KeychainItemWrapper alloc] initWithIdentifier:LITTLE_NOTE_IDENTIFIER accessGroup:nil];
    [keychain resetKeychainItem];
    [NSUserDefaults resetStandardUserDefaults];
    [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kUserLogoutNotification
                                                                                        object:nil]];
    
}

-(BOOL)isSharing{
    return _personSharingWith != nil;
}

-(NSString *)sharingUsername{
    return [NSString stringWithString:_personSharingWith];
}

-(void)stopSharing{
    dispatch_async(dispatch_get_main_queue(), ^{
        [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:YES];
    });
    if(!_csrfToken){
        [self getCookie:^{
            [self stopSharing];
        }];
    }else{
        NSString *urlString = [URLFeeder getURLToBreakConnection];
        urlString = [urlString stringByAppendingString:@"?connection=0"];
        NSURL *url = [NSURL URLWithString:urlString];
        NSMutableURLRequest *request = [self getLittleNoteRequestForURL:url];
        [request setHTTPMethod:@"PUT"];
        _personSharingWith = nil;
        [NSURLConnection sendAsynchronousRequest:request
                                           queue:[[NSOperationQueue alloc]init]
                               completionHandler:^(NSURLResponse *response, NSData *data, NSError *error){
                                   if(error){
                                       [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kConnectionFailedNotification
                                                                                                                           object:nil]];
                                   }else{
                                       NSHTTPURLResponse *httpRespone = (NSHTTPURLResponse *)response;
                                       if([httpRespone statusCode] == 200){
                                           [[NoteManager getInstance] deleteAll];
                                           [[NSUserDefaults standardUserDefaults] setObject:nil forKey:SHARE_WITH_KEY];
                                           [[NSUserDefaults standardUserDefaults] synchronize];
                                           [[NSNotificationCenter defaultCenter]postNotification:[NSNotification notificationWithName:kStopSharingSuccussfulNotification
                                                                                                                               object:nil]];
                                       }
                                   }
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                       [[UIApplication sharedApplication]setNetworkActivityIndicatorVisible:NO];
                                   });
                               }];
    }

}


@end
